#coding: utf-8
from store import Storage, PackageLoader, Meta, Profile, Remote
import sys
import optparse
import json
import os
import getpass
import uuid

def _cmd_remote(argv):
    p = optparse.OptionParser("%s [options] [remote]" % argv[0])

    p.add_option('-i', '--rest-url', 
            action="store", type="string", dest="rest_url", default="",
            help="set base url to REST API for remote setting")

    p.add_option('-a', '--rest-ca', 
            action="store", type="string", dest="rest_ca", default="",
            help="set CAfile to trust REST for remote setting")

    try:
        opts, args = p.parse_args(argv[1:])
    except:
        p.print_help(sys.stdout)
        return

    conn = Storage().conn

    if len(args)==0:
        for x in Remote.list(conn):
            if x['pri']==1:
                pri = "*"
            else:
                pri = " "
            sys.stdout.write("{} {}\t{}\t{}\n".format(pri, x['remote_id'], x['rest_url'], x['rest_ca']!=""))
        return
        
    remote = Remote.find(conn, args[0])
    if not remote:
        remote = Remote(remote_id=args[0])

    if opts.rest_url!="":
        remote['rest_url'] = opts.rest_url

    if opts.rest_ca=="none":
        remote['rest_ca'] = ""
    elif opts.rest_ca!="":
        fh = open(opts.rest_ca)
        t = fh.read()
        fh.close()
        remote['rest_ca'] = t

    if remote.get("rest_url", "")=="":
        sys.stderr.write("invalid REST API url\n")
        sys.exit(-1)

    remote.save(conn)
    conn.commit()
    conn.close()


def _cmd_install(argv):
    p = optparse.OptionParser("%s [options] package-path" % argv[0])

    try:
        opts, args = p.parse_args(argv[1:])
        assert len(args)>0
    except:
        p.print_help(sys.stdout)
        return

    conn = Storage().conn
    for path in args:
        if not os.path.exists(path):
            sys.stderr.write("path '{}' not exists\n".format(path))
            continue

        pkg = PackageLoader(path)
        for index, (modname, cls) in pkg.classes.items():
            meta = Meta(**{
                "meta_id": index, 
                "path": path, 
                "module": modname, 
                "class": cls.__name__,
                "name": cls.NAME,
                "pid": cls.PID
            })
            meta.save(conn)
            sys.stdout.write("installed %s\n" % index)
    conn.commit()
    conn.close()


def _cmd_list(argv):
    conn = Storage().conn
    for meta in Meta.list(conn):
        sys.stdout.write("%s\n" % meta['meta_id'])
        for p in Profile.list(conn, meta["meta_id"]):
            sys.stdout.write("\t{}\t{}\n".format(p['name'], p['remote_id']))
    conn.close()


def _cmd_register(argv):
    p = optparse.OptionParser("%s [options] thing" % argv[0])

    p.add_option('-r', '--remote', 
            action="store", type="string", dest="remote", default="",
            help="remote settings, default to the primary one")

    p.add_option('-t', '--token', 
            action="store", type="string", dest="token", default="",
            help="identity token of thing, auto-generate one if not given")

    p.add_option('-o', '--profile',
            action="store", type="string", dest="profile", default="",
            help="profile name that used to store register results")

    p.add_option('-f', '--force',
            action="store_true", dest="force", default=False,
            help="force to override if profile name is existed")

    p.add_option("-U", "--username", 
            action="store", type="string", dest="username", default="",
            help="username of your account")

    p.add_option("-P", "--password", 
            action="store", type="string", dest="password", default="",
            help="password your account, prompt to ask if omitted")

    try:
        opts, args = p.parse_args(argv[1:])
        assert len(args)>0
    except:
        p.print_help(sys.stdout)
        sys.exit(-1)

    conn = Storage().conn

    # check thing
    meta = Meta.find(conn, args[0])
    if not meta:
        if os.path.exists(args[0]):
            pkg = PackageLoader(args[0])
            if len(pkg.classes)>1:
                sys.stderr.write("more than one thing found in {}\n".format(args[0]))
                for k in pkg.classes:
                    sys.stderr.write("    {}\n".format(k))
                sys.stderr.write("don't known which one to register, abort.\n".format(args[0]))
                sys.exit(-1)
            elif len(pkg.classes)==0:
                sys.stderr.write("no thing class found in {}, abort.\n".format(args[0]))
                sys.exit(-1)
            else:
                meta = Meta.find(conn, pkg.classes.keys()[0])
                if meta==None:
                    sys.stdout.write("thing '{}' is not installed yet.\nlet's install it first ...".format(pkg.classes.keys()[0]))
                    meta_id, (modname, cls) = pkg.classes.items()[0]
                    meta = Meta(**{
                        "meta_id": meta_id, 
                        "path": pkg.path, 
                        "module": modname, 
                        "class": cls.__name__,
                        "name": cls.NAME,
                        "pid": cls.PID
                    })
                    meta.save(conn)
                    sys.stdout.write(" done\n")

    if not meta:
        sys.stderr.write("unkown thing {}\n".format(args[0]))
        sys.exit(-1)

    meta_id = meta._id
    cls = meta.get_class()
    if cls==None:
        sys.stderr.write("can't load thing {}\n".format(meta_id))
        sys.exit(-1)

    # check profile
    profile = opts.profile
    if profile == "":
        profile_id = meta_id
    else:
        for c in "*/?`()[]{}^&~;.":
            if c in opts.profile:
                sys.stderr.write("profile name contains invalid char %s\n" % c)
                sys.exit(-1)
        profile_id = "{}/{}".format(meta_id, opts.profile)

    prf = Profile.find(conn, profile_id)
    if prf:
        if not opts.force:
            sys.stderr.write("profile %s already existed\n" % profile_id)
            sys.exit(-1)

        # use existing profile attrs
        if prf['meta_id']!=meta_id:
            sys.stderr.write(
                    "thing name mismatch: '{}' vs '{}' in profile {}\n"
                    .format(meta_id, prf['meta_id'], profile_id)
            )
            sys.exit(-1)

        if opts.token=="":
            opts.token = prf['token']
        elif opts.token!=prf['token']:
            sys.stderr.write("token mismatch\n")
            sys.exit(-1)

        if opts.remote=="":
            opts.remote = prf['remote_id']
        elif opts.remote!=prf['remote_id']:
            sys.stderr.write("remote mismatch\n")
            sys.exit(-1)

    # check for remote
    if opts.remote=="":
        remote = Remote.primary(conn)
        if not remote:
            sys.stderr.write("no primary remote setting\n")
            sys.exit(-1)
    else:
        remote = Remote.find(conn, opts.remote)
        if not remote:
            sys.stderr.write("unkown remote setting {}\n".format(opts.remote))
            sys.exit(-1)

    if opts.token!="":
        if "/" in opts.token:
            print(sys.stderr, "char '/' is invalid for token")
            sys.exit(-1)
    else:
        opts.token = str(uuid.uuid1()).replace("-", "")

    secret = str(uuid.uuid1()).replace("-", "")

    if opts.username=="":
        sys.stderr.write("username is required\n")
        sys.exit(-1)

    passwd = opts.password
    while passwd=="":
        try:
            passwd = getpass.getpass("password for %s: " % opts.username)
        except KeyboardInterrupt:
            sys.stdout.write("\n")
            sys.exit(-1)

    # print register info
    sys.stdout.write("registering:\n")
    sys.stdout.write("    thing:   {}\n".format(meta_id))
    sys.stdout.write("    token:   {}\n".format(opts.token))
    sys.stdout.write("    remote:  {} ({})\n".format(remote._id, remote['rest_url']))
    sys.stdout.write("    owner:   {}\n".format(opts.username))
    sys.stdout.write("    profile: {}\n".format(profile_id))

    # make thing instance & register
    thing = cls(opts.token, secret=secret)
    thing.set_rest(remote['rest_url'], remote['rest_ca'])
    status, content = thing.register(opts.username, passwd)
    if status != 200:
        sys.stderr.write("register fail ({}): {}\n".format(status, content))
        sys.exit(-1)

    ack = None
    try:
        d = json.loads(content)
    except:
        sys.stderr.write("invalid response message %s\n" % content)
        sys.exit(-1)

    if d.get('status') != 0:
        sys.stderr.write("ERROR(%d): %s\n" % (d['status'], d.get('response',"")))
        sys.exit(-1)

    ack = d['response']

    # save profile
    pfile = Profile(profile_id=profile_id, meta_id=meta_id, name=opts.profile, token=opts.token, \
            password=secret, remote_id=opts.remote, owner=opts.username)

    if "cert" in ack:
        pfile['iot_cert'] = ack['cert'][0]
        pfile['iot_key'] = ack['cert'][1]
    else:
        pfile["iot_cert"] = ""
        pfile["iot_key"] = ""

    if "trust" in ack:
        pfile['iot_ca'] = ack['trust']
    else:
        pfile['iot_ca'] = ""

    pfile.save(conn)
    conn.commit()
    conn.close()

    sys.stdout.write("successfully registered as '{}'\n".format(profile_id))


def _cmd_start(argv):
    p = optparse.OptionParser("%s [options] <thing> <profile>" % argv[0])

    try:
        opts, args = p.parse_args(argv[1:])
        assert len(args)>=1
    except:
        p.print_help(sys.stdout)
        sys.exit(-1)

    conn = Storage().conn

    profile = Profile.find(conn, args[0])
    if not profile:
        sys.stderr.write("can't load profile {}\n".format(args[0]))
        sys.exit(-1)

    profile.show_info(sys.stdout)
    thing = profile.make_thing(conn)
    if thing==None:
        sys.stderr.write("can't load make thing for {}\n".format(args[0]))
        sys.exit(-1)

    status, resp = thing.load_remote_profile()
    if status == 0:
        conn.close()
        thing.start()
    elif status == 404:
        sys.stderr.write("can't find register info on host, clear registry for '%s'\n" % profile)
        _profile_clear_registry(profile)
    else:
        sys.stderr.write("cannot load profile from host: %s %s\n" % (status, resp))
        sys.exit(-1)


def commands():
    p = optparse.OptionParser(usage="%prog cmd [options]")

    p.disable_interspersed_args()
    opts, args = p.parse_args()

    if len(args)==0:
        p.print_help(sys.stderr)
        sys.stderr.write("\nwhere cmd may be: install, list, remote, register, start\n")
        sys.exit(-1)

    cmd = args[0]
    if cmd=="install":
        _cmd_install(args)
    elif cmd=="list":
        _cmd_list(args)
    elif cmd=="remote":
        _cmd_remote(args)
    elif cmd=="register":
        _cmd_register(args)
    elif cmd=="start":
        _cmd_start(args)
    else:
        sys.stderr.write("unknown command %s\n" % cmd)
        p.print_help(sys.stderr)
        sys.exit(-1)




#coding: utf-8
from nodewox.thing import NX_PREFIX
from param import Param
import types
import collections
import sys
import traceback
import struct

class Node(object):
    "node base class"
    NAME = "" # default node name

    def __init__(self, key, name="", parent=None, comment="", **kwargs):
        assert isinstance(key, basestring), key
        assert key!="" and "/" not in key, key

        if name=="":
            name = self.NAME

        if parent!=None:
            assert isinstance(parent, Node), parent

        self._key = key
        self._name = name
        self._parent = parent
        self._comment = comment
        self._params = {}
        self._children = {}
        self._seq = 0
        self._id = None

        self.setup()


    @property
    def key(self): 
        return self._key

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        return self._children

    def setup(self):
        pass

    def get_id(self):
        return self._id

    def add_child(self, node, seq=0):
        assert isinstance(node, Node), node
        assert node._parent==self
        assert node.key not in self._children, node.key
        if node._seq==0:
            node._seq = len(self._children)+1
        self._children[node.key] = node
        return node


    #
    # NODE PARAMS RELATED
    #
    def has_param(self, key):
        return self._params.has_key(key)

    def add_param(self, key, value, name="", comment="", options=[], flag="volatile", seq=-1):
        "defina a parameter with type of value"
        assert not self.has_param(key), key
        p = Param(key, value, name=name, options=options, flag=flag, comment=comment, seq=seq)
        if seq<0: p.seq = len(self._params) + 1
        self._params[key] = p
        return p

    def set_param(self, key, value):
        "set new value to a param"
        if key in self._params:
            return self._params[key].set_value(value)
        else:
            return False

    def set_params(self, vals):
        "set values of many params"
        assert type(vals)==types.DictType, vals
        cnt = 0
        for k, v in vals.items():
            if self.set_param(k, v):
                cnt += 1
        return cnt

    def get_param(self, key, default=None):
        if key not in self._params:
            return default
        else:
            return self._params[key].value

    def reset_params(self):
        "reset all params to itsinitial value"
        for p in self._params.values():
            p.reset()


    def as_data(self):
        assert isinstance(self._key, basestring) and self._key!=""
        res = {"key":self._key, "name":self._name, "seq":self._seq}

        if self._comment!="":
            res['comment'] = self.comment

        if len(self._params)>0:
            res['params'] = dict((k,v.as_data()) for k,v in self._params.items())

        return res


    def handle_request(self, action="", params={}, children=[]):
        "processe an incomming request"
        assert self._id > 0
        report_params = (action=="status")

        # set params
        for k,v in params.items():
            p = self._params.get(k)
            if p!=None:
                oval = p.value 
                if self.set_param(k, v):
                    self.on_param_changed(p, old_value=oval)
                    report_params = True

        # resport param status
        res = {}
        if report_params and len(self._params)>0:
            params = {}
            for p in self._params.values():
                if p.flag != "static":
                    params[p.key] = p.value
            if len(params)>0:
                res['params'] = params

        pubs = {"{}{}/r".format(NX_PREFIX, self.get_id()): res}
        if len(children)>0:
            # request into children
            for c in self.children.values():
                if c.get_id() in children:
                    r = c.handle_request(action=action)
                    if type(r)==types.DictType:
                        pubs.update(r)
        return pubs


    def loop(self):
        pass

    def on_param_changed(self, param, old_value):
        pass


    @classmethod
    def decode_packet(cls, packet, datatype, dim):
        assert isinstance(packet, (bytearray, basestring)), packet
        assert isinstance(datatype, basestring), datatype
        assert dim>=0, dim

        if datatype=="":
            data = bytearray(packet)
        elif len(packet)==0:
            data = []
        else:
            try:
                if datatype=="byte":
                    n = len(packet) / struct.calcsize("B")
                    data = struct.unpack("%dB" % n, packet)
                elif datatype=="int16":
                    n = len(packet) / struct.calcsize("h")
                    data = struct.unpack("!%dh" % n, packet)
                elif datatype=="int32":
                    n = len(packet) / struct.calcsize("i")
                    data = struct.unpack("!%di" % n, packet)
                elif datatype=="int64":
                    n = len(packet) / struct.calcsize("q")
                    data = struct.unpack("!%dq" % n, packet)
                elif datatype=="float":
                    n = len(packet) / struct.calcsize("f")
                    data = struct.unpack("!%df" % n, bytearray(packet))
                elif datatype=="bool":
                    n = len(packet) / struct.calcsize("B")
                    data = tuple(x!=0 for x in struct.unpack("!%dB" % n, packet))
                elif datatype=="string":
                    data = []
                    start = 0
                    while start < len(packet):
                        sz = packet[start:].find('\0')
                        assert sz>=0
                        if sz==0:
                            data.append("")
                        else:
                            data.append(struct.unpack("%ds" % sz, packet[start:start+sz]))
                        start += sz+1
                else:
                    raise NotImplementedError

                data = list(data)

            except:
                data = None
                traceback.print_exc(file=sys.stderr)

        if isinstance(data, list) and dim>0 and len(data)<dim:
            # padding list
            n = dim - len(data)
            if datatype in ("byte", "int16", "int32", "int64", "float"):
                data += [0] * n
            elif datatype=="bool":
                data += [False] * n
            elif datatype=="string":
                data += [""] * n

        return data


    @classmethod
    def encode_packet(cls, data, datatype, dim):
        if datatype == "":
            # string and bytearray converts to bytes
            # other data converts to null
            p = ""
            if data is not None and len(data)>0:
                if isinstance(data, basestring):
                    data = bytearray(data)
                    p = struct.pack("%dB" % len(data), *data)
                elif isinstance(data, bytearray):
                    p = struct.pack("%dB" % len(data), *data)

        else:
            assert isinstance(data, (list, tuple)), data

            if datatype=="byte":
                p = struct.pack("%dB" % len(data), *data)
            elif datatype=="int16":
                p = struct.pack("!%dh" % len(data), *data)
            elif datatype=="int32":
                p = struct.pack("!%di" % len(data), *data)
            elif datatype=="int64":
                p = struct.pack("!%dq" % len(data), *data)
            elif datatype=="float":
                p = struct.pack("!%df" % len(data), *data)
            elif datatype=="bool":
                p = struct.pack("!%db" % len(data), *data)
            elif datatype=="string":
                fmt = ""
                for s in data:
                    fmt += "%ds" % (len(s)+1)
                p = struct.pack(fmt, *data)
            else:
                raise NotImplementedError

        return bytearray(p)


    @classmethod
    def check_data(cls, data, datatype, dim):
        if datatype=="":
            if isinstance(data, basestring):
                data = bytearray(data)
            assert isinstance(data, bytearray), data

        else:
            if isinstance(data, bytearray):
                data = list(data)
            elif not isinstance(data, (list, tuple)):
                data = [data]
            else:
                if dim > 0 and len(data) < dim:
                    data = data[:dim] # truncate

            # fix data value
            if datatype=="byte":
                data = map(_byte, data)
            elif datatype=="int16":
                data = map(_int16, data)
            elif datatype=="int32":
                data = map(_int32, data)
            elif datatype=="int64":
                data = map(_int64, data)
            elif datatype=="float":
                data = map(_float, data)
            elif datatype=="bool":
                data = map(_bool, data)
            elif datatype=="string":
                data = map(_string, data)
            else:
                assert False, (datatype, data)

        assert isinstance(data, (list, tuple, bytearray)), data
        return data


#
# helpers
#
def _int(v):
    r = 0
    try:
        if isinstance(v, int):
            r = v
        elif isinstance(v, basestring):
            r = int(v)
        elif isinstance(v, bool):
            r = int(bool)
        elif isinstance(v, float):
            r = int(v)
    except:
        r = 0
    return r

def _byte(v):
    return _int(v) & 0xFF

def _int16(v):
    return _int(v) & 0xFFFF

def _int32(v):
    return _int(v) & 0xFFFFFFFF

def _int64(v):
    return _int(v) & 0xFFFFFFFFFFFFFFFF

def _float(v):
    r = 0
    try:
        if isinstance(v, float):
            r = v
        elif isinstance(v, int):
            r = float(v)
        elif isinstance(v, basestring):
            r = float(v)
        elif isinstance(v, bool):
            r = int(bool)
    except:
        r = 0
    return r

def _string(v):
    r = ""
    if isinstance(v, basestring):
        r = v
    elif isinstance(v, (int, float, bytearray)):
        r = str(v)
    elif isinstance(v, bool):
        r = str(bool).lower()
    return r

def _bool(v):
    r = False
    if isinstance(v, bool):
        r = v
    elif isinstance(v, (int, float)):
        r = int(v)!=0
    elif isinstance(v, basestring):
        r = v.lower() in ("true", "t", "yes", "y", "on", "1")
    return r


def U8(data):
    "convert data to utf-8"
    if isinstance(data, collections.Mapping):
        return dict((U8(key),U8(value)) for key, value in data.iteritems())
    elif isinstance(data, list):
        return [U8(x) for x in data]
    elif isinstance(data, unicode):
        return data.encode('utf-8')
    else:
        return data




#coding: utf-8
from nodewox.thing import NX_PREFIX
from node import Node
from Queue import Queue
import types
import time

class Channel(Node):

    # channel datatype decl.
    DATA_TYPE = ("", 0)

    def __init__(self, thing, key, gender, name="", latch=False, seq=0, exclusive=True, comment="", **kwargs):
        from thing import Thing
        assert isinstance(thing, Thing), thing
        assert gender in ("F", "M"), gender

        assert self.DATA_TYPE[0] in ("int16", "int32", "int64", "byte", "float", "bool", "string", ""), self.DATA_TYPE
        assert self.DATA_TYPE[1]>=0, self.DATA_TYPE

        self._parent = thing
        self._gender = gender
        self._latch = latch
        self._exclusive = exclusive
        self._seq = seq
        self._time_wakeup = 0

        if name=="": name = self.NAME
        if name=="": name = key

        Node.__init__(self, key, parent=thing, name=name, comment=comment, **kwargs)


    @property
    def is_awake(self):
        if self._time_wakeup!=0 and time.time() >= self._time_wakeup:
            self._time_wakeup = 0
        return self._time_wakeup==0

    def sleep(self, ms):
        if ms > 0:
            self._time_wakeup = max(self._time_wakeup, time.time() + ms/1000.0)

    def add_child(self, node):
        raise NotImplementedError

    def as_data(self):
        assert self._parent!=None
        res = Node.as_data(self)
        res.update({
                "gender": self._gender,
                "latch": self._latch,
                "lang": "",
                "seq": self._seq,
        })

        if self.DATA_TYPE[0]=="":
            res['datatype'] = ("", 0)
        else:
            res['datatype'] = self.DATA_TYPE

        if self._gender=="F":
            res['exclusive'] = self._exclusive
        else:
            res['exclusive'] = False

        return res


class MaleChannel(Channel):
    DATA_TYPE = ("", 0)
    QSIZE = 10

    def __init__(self, thing, key, **kwargs):
        Channel.__init__(self, thing, key, "M", **kwargs)
        self._dataq = Queue(self.QSIZE)

    def clear_data(self):
        while not self._dataq.empty():
            self._dataq.get()

    def feed_data(self, data, gid=0, src=0):
        data = self.check_data(data, self.DATA_TYPE[0], self.DATA_TYPE[1])
        if self._dataq.full(): 
            self._dataq.get()
        self._dataq.put((data, gid, src))

    def send_data(self):
        n = 0
        if self.get_id()>0:
            mess = self.parent.get_messenger()
            if mess.is_connected:
                topic = "{}{}".format(NX_PREFIX, self.get_id())
                while not self._dataq.empty():
                    data, gid, src = self._dataq.get()
                    p = self.encode_packet(data, self.DATA_TYPE[0], self.DATA_TYPE[1])
                    assert isinstance(p, bytearray), p
                    mess.publish(topic, p)
                    n += 1
        return n


    def loop(self):
        self.send_data()


class FemaleChannel(Channel):
    DATA_TYPE = ("", 0)

    def __init__(self, thing, key, **kwargs):
        Channel.__init__(self, thing, key, "F", **kwargs)

    def handle_packet(self, packet, src=0):
        data = self.decode_packet(packet, self.DATA_TYPE[0], self.DATA_TYPE[1])
        if data!=None:
            return self.perform(data, src=src)  # perform action

    def perform(self, data, gid=0, src=0):
        raise NotImplementedError




#coding: utf-8
from nodewox.thing import NX_PREFIX
from node import Node, U8
from messenger import Messenger
from channel import Channel
import signal
import json
import types
import os
import re
from M2Crypto import SSL, RSA, X509, m2, httpslib
import urlparse

PAT_REQ   = re.compile(r"^{}(\d+)\/q$".format(NX_PREFIX.replace("/", r"\/")))
PAT_EVENT = re.compile(r"^{}(\d+)$".format(NX_PREFIX.replace("/", r"\/")))

class Thing(Node):
    # thing prop decl.
    NAME = ""
    PID = None
    CHKEYS = "*"

    # messenger class decl.
    MESSENGER_CLASS = Messenger

    def __init__(self, key, secret="", reconnect=1000):
        assert isinstance(key, basestring) and key!="" and "/" not in key, key
        assert issubclass(self.MESSENGER_CLASS, Messenger), self.MESSENGER_CLASS

        self._key = key
        self._secret = secret
        self._rest_url = ""
        self._rest_ca = ""

        self._reconnect = reconnect
        
        self._cafile = None
        self._certfile = None
        self._keyfile = None

        self._username = key
        self._secret = secret

        # runtime
        self._messenger = None
        self._running = False

        Node.__init__(self, self._key, name=self.NAME)


    @property
    def is_registered(self):
        return self._certfile!="" and self._rest_url!=""

    def set_rest(self, url, ca=""):
        self._rest_url = url
        self._rest_ca = ca

    def set_user_pw(self, username, password):
        self._username = username
        self._secret = password

    def set_tls(self, cafile, certfile, keyfile):
        self._cafile = cafile
        self._certfile = certfile
        self._keyfile = keyfile

    def get_messenger(self):
        if self._messenger==None:
            assert self._host!=""
            args = {
                    "host": self._host,
                    "port": self._port,
                    "unpw": (self._username, self._secret),
                    "cafile": self._cafile,
                    "certfile": self._certfile,
                    "keyfile": self._keyfile,
                    "reconnect": self._reconnect,
            }
            self._messenger = self.MESSENGER_CLASS(self, **args)
        assert isinstance(self._messenger, Messenger), self._messenger
        return self._messenger


    def add_child(self, ch):
        assert isinstance(ch, Channel), ch
        return Node.add_child(self, ch)


    def as_data(self):
        res = Node.as_data(self)
        if self.PID==None:
            # user defined
            assert len(self.children)>0
        else:
            # instantiate from product
            assert type(self.PID)==types.IntType, self.PID
            res['product'] = self.PID

        res['children'] = dict((x.key,x.as_data()) for x in self.children.values())
        return res


    def _make_ssl_context(self):
        return SSL.Context()

    def load_remote_profile(self):
        "从主机获取节点配置"
        assert self._rest_url!=""
        assert self.is_registered

        headers = {
            "authorization": "CERT %s" % self._secret,
            "x-requested-with": "XMLHttpRequest",
        }

        sctx = self._make_ssl_context()
        if self._certfile not in (None, ""):
            cert = X509.load_cert_string(self._certfile)
            key = RSA.load_key_string(self._keyfile)
            m2.ssl_ctx_use_x509(sctx.ctx, cert.x509)
            m2.ssl_ctx_use_rsa_privkey(sctx.ctx, key.rsa)

        u = urlparse.urlparse(os.path.join(self._rest_url, "thing/profile"))
        conn = httpslib.HTTPSConnection(u.netloc, ssl_context=sctx)
        conn.request("GET", u.path, headers=headers)

        res = conn.getresponse()
        status = res.status
        content = res.read()

        if status != 200:
            return status, content

        try:
            d = U8(json.loads(content))
            if d['status'] != 0:
                return d['status'], d['response']

            resp = U8(d['response'])

            # thing id
            self._id = resp['id']
            assert type(self._id)==types.IntType, self._id
            
            # mqtt config
            p = urlparse.urlsplit(resp['mqtt'])
            a = p.netloc.split(":")
            if len(a)>1:
                host = a[0]
                port = int(a[1])
            else:
                host = a[0]
                if p.scheme in ("mqtts", "ssl"):
                    port = 8883
                else:
                    port = 1883
            self._host = host
            self._port = port

            # setup children
            for k, chinfo in resp["children"].items():
                print(k, chinfo['id'])

                if k in self.children:
                    ch = self.children[k]
                    assert ch._gender==chinfo['gender']

                    ch._id = chinfo['id']
                    assert type(ch._id)==types.IntType, ch._id

                    for field in ("datatype", "exclusive", "latch"):
                        if field in chinfo:
                            setattr(ch, "_%s" % field, chinfo[field])

                    # setup params
                    params = chinfo.get("params")
                    if params!=None:
                        pvs = {}
                        for pkey, pinfo in params.items():
                            if not ch.has_param(pkey):
                                ch.add_param(pkey, **pinfo)

                            if pinfo.get("value")!=None:
                                pvs[pkey] = pinfo['value']

                        ch.reset_params()
                        if len(pvs)>0:
                            ch.set_params(pvs)

            return 0, resp

        except:
            import traceback
            print("*" * 20)
            traceback.print_exc()
            print("*" * 20)
            return -1, "ERROR"


    def register(self, user, passwd):
        "register the thing to nodewox host"
        assert self._secret not in (None, "")
        assert self._rest_url!=""

        # make thing meta
        info = self.as_data()
        info['secret'] = self._secret

        headers = {
                "authorization": "USERPW %s\t%s" % (user, passwd),
                "x-requested-with": "XMLHttpRequest",
                "content-type": "application/json",
        }
        u = urlparse.urlparse(os.path.join(self._rest_url, "thing/register?trust=pem&cert=pem"))
        conn = httpslib.HTTPSConnection(u.netloc, ssl_context=self._make_ssl_context())
        conn.request("POST", "{}?{}".format(u.path, u.query), body=U8(json.dumps(info)), headers=headers)

        res = conn.getresponse()
        status = res.status
        content = res.read()
        return status, content


    def handle_request(self, action="", params={}, children=[]):
        if action=="unregister":
            self.on_unregister()
        else:
            return super(Thing, self).handle_request(action=action, params=params, children=children)

    def on_unregister(self):
        self.stop()

    def on_connected(self, userdata):
        # listen on topics
        res = {}
        mess = self.get_messenger()

        subs = ["{}{}/q".format(NX_PREFIX, self._id)]
        for ch in self.children.values():
            assert type(ch._id)==types.IntType and ch._id>0, (ch.key, ch._id)
            subs.append("{}{}/q".format(NX_PREFIX, ch._id))
            if ch._gender=="F":
                subs.append("{}{}".format(NX_PREFIX, ch._id))

            # say hello from channel ch
            res2 = ch.handle_request(action="status")
            if res2!=None:
                res.update(res2)

        if len(self._params)>0:
            # say hello from this thing
            res2 = self.handle_request(action="status")
            if res2!=None:
                res.update(res2)

        # sub & pub
        mess.subscribe(subs)
        for topic, msg in res.items():
            assert isinstance(msg, dict), msg
            if len(msg)>0:
                payload = bytearray(json.dumps(msg))
            else:
                payload = ""
            mess.publish(topic, payload, qos=0)


    def on_connect_fail(self, code, userdata):
        pass

    def on_disconnected(self, code, userdata):
        pass

    def on_message(self, msg, userdata):
        pass

    def on_sig(self, *args):
        print("SIG", args[0])
        self._running = False


    def loop(self):
        # tick each channel
        for ch in self.children.values():
            if ch.is_awake:
                ch.loop()


    def start(self):
        assert self.is_registered
        assert not self._running

        signal.signal(signal.SIGINT, self.on_sig)
        signal.signal(signal.SIGTERM, self.on_sig)

        # main loop
        mess = self.get_messenger()
        self._running = True

        while self._running:
            self.loop() # drive thing work
            mess.loop() # drive messenger work


    def stop(self):
        self._running = False




from setuptools import setup
from setuptools.command.install import install
import subprocess
import os

NODEWOX_HOME = "/var/lib/nodewox"

'''
class NXInstallCommand(install):

    def run(self):
        # make NODEWOX_HOME
        if not os.path.exists(NODEWOX_HOME):
            print("make dir %s" % NODEWOX_HOME)
            os.mkdir(NODEWOX_HOME, 0o755)

        for x in ("trust", "profiles"):
            p = os.path.join(NODEWOX_HOME, x)
            if not os.path.exists(p):
                print("make dir %s" % p)
                os.mkdir(p, 0o755)

        # copy default cacerts
        subprocess.check_output([
            'cp', 'data/cacert.pem', os.path.join(NODEWOX_HOME, "trust/default.pem")
        ])

        install.run(self)
'''

setup(
	name="nodewox-pything",
	version="0.1.1",
	author="johnray",
	author_email="996351336@qq.com",
	description="nodewox sdk for things",
        license="MIT",

        packages=["nodewox.thing"],
        #cmdclass={'install': NXInstallCommand},

        install_requires=["nodewox-mqtt>=1.2", "M2Crypto>=0.25.1"],

        entry_points = """
        [console_scripts]
        nx_thing = nodewox.thing.cmd_thing:commands
        """,

        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Environment :: Console',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Topic :: Internet',
            'Topic :: Education',
        ]        
)




#coding: utf-8
import json

class Param(object):

    def __init__(self, key, value, name="", flag="volatile", options=[], comment="", seq=0):
        assert key not in ("", "id"), key
        assert flag in ("volatile", "persistent", "readonly", "static"), flag

        # check validity of init value
        self.datatype = None
        if value in (True, False):
            self.datatype = bool
        else:
            for t in (int, float, basestring):
                if isinstance(value, t):
                    self.datatype = t
                    break
        if self.datatype==None:
            raise Exception("invalid init value: must be a value of int, float, bool or string type")

        self.key = key
        self.name = name
        self.flag = flag
        self.init_value = value
        self.value = value
        self.seq = seq
        self.comment = comment
        self.disabled = False

        if len(options)>0:
            # check options
            assert isinstance(options, (list,type)), options
            assert self.datatype in (int, basestring)
            for i, x in enumerate(options):
                if isinstance(x, (list,tuple)):
                    assert isinstance(x[0], self.datatype)
                    assert isinstance(x[1], basestring)
                else:
                    assert isinstance(x, basestring)
                    if self.datatype==int:
                        val = i
                    else:
                        val = str(i)
                    options[i] = (val, x)
        self.options = options


    def set_value(self, val):
        if self.flag in ("volatile", "persistent", "readonly"):
            if self.datatype==float and isinstance(val, int):
                val = float(val)
            elif self.datatype==int and isinstance(val, (float,bool)):
                val = int(val)
            elif self.datatype==bool and isinstance(val, (int,float)):
                val = val!=0

            if isinstance(val, self.datatype):
                self.value = val
                return True

        return False


    def reset(self):
        self.value = self.init_value


    def as_data(self):
        res = {
            "key": self.key,
            "value": self.value,
            "flag": self.flag,
            "seq": self.seq,
        }

        if self.datatype==int:
            res['datatype'] = "int"
        elif self.datatype==float:
            res['datatype'] = "float"
        elif self.datatype==basestring:
            res['datatype'] = "string"
        elif self.datatype==bool:
            res['datatype'] = "bool"
        else:
            raise NotImplementedError

        if self.name not in ("", self.key):
            res['name'] = self.name

        if self.comment!="":
            res['comment'] = self.comment

        if len(self.options)>0:
            res['options'] = options

        return res




#coding: utf-8
from nodewox.thing import NX_PREFIX
from nodewox.mqtt.client import Client
import time
import traceback
import json
import re

class Messenger(object):
    PAT_REQ    = re.compile(r"^{}(\d+)\/q$".format(NX_PREFIX.replace("/", r"\/")))
    PAT_RES    = re.compile(r"^{}(\d+)\/r$".format(NX_PREFIX.replace("/", r"\/")))
    PAT_PACKET = re.compile(r"^{}(\d+)$".format(NX_PREFIX.replace("/", r"\/")))

    def __init__(self, master_node, host="", port=-1, unpw=None, certfile=None, keyfile=None, cafile=None, reconnect=1000):
        from node import Node
        assert isinstance(master_node, Node), master_node

        self._node = master_node
        self._host = host
        self._port = port
        self._reconnect = reconnect

        self._certfile = certfile
        self._keyfile = keyfile
        self._cafile = cafile

        if unpw!=None:
            self._username, self._password = unpw
        else:
            self._username = self._password = ""

        # runtime
        self._reconnect_factor = 1
        self._reconnect_try = 0
        self._next_connect_time = time.time()
        self._connected = False
        self._client = None


    @property
    def is_connected(self):
        return self._client!=None and self._connected

    def get_reconnect_interval(self):
        return 1000

    def _sched_next_connect(self):
        timeout = self.get_reconnect_interval()
        if timeout > 0:
            if self._reconnect_factor<10:
                self._reconnect_factor = 1 + self._reconnect_try // 100
            print("schedule to reconnect after %sms" % (timeout * self._reconnect_factor))
            self._next_connect_time = time.time() + timeout * self._reconnect_factor/1000.0


    def ack_msg_request(self, client, userdata, msg):
        _id = int(self.PAT_REQ.findall(msg.topic)[0])
        if _id == self._node.get_id():
            target = self._node
        else:
            target = None
            for ch in self._node.children.values():
                if ch.get_id()==_id:
                    target = ch
                    break

        if target != None:
            req = {}
            if len(msg.payload)>0:
                try:
                    req = json.loads(msg.payload)
                except:
                    traceback.print_exc()
                    print("ERROR: invalid request message")
                    return

            args = {
                    "action": req.get("action", ""),
                    "params": req.get("params") or {},
                    "children": req.get("children") or [],
            }
            res = target.handle_request(**args)

        if isinstance(res, dict):
            for topic, msg in res.items():
                assert isinstance(msg, dict), msg
                if len(msg)==0:
                    payload = ""
                else:
                    payload = bytearray(json.dumps(msg))
                client.publish(topic, payload)


    def ack_msg_packet(self, client, userdata, msg):
        _id = int(self.PAT_PACKET.findall(msg.topic)[0])
        chs = [x for x in self._node.children.values() if x.get_id()==_id]
        if len(chs) > 0:
            chan = chs[0]
            if hasattr(chan, "handle_packet"):
                chan.handle_packet(bytearray(msg.payload))


    def _on_connect(self, client, userdata, flags, rc):
        if rc==0:
            self._connected = True
            self._reconnect_try = 0
            self._next_connect_time = 0
            self._node.on_connected(userdata)
        else:
            self._connected = False
            self._node.on_connect_fail(rc, userdata)
            self._sched_next_connect()

    def _on_disconnect(self, client, userdata, rc):
        self._connected = False
        self._node.on_disconnected(rc, userdata)
        if client!=None:
            self._sched_next_connect()

    def _on_subscribe(self, client, userdata, mid, granted_qos):
        pass

    def _on_message(self, client, userdata, msg):
        self._node.on_message(msg, userdata)

    def _on_log(self, client, userdata, level, buf):
        print(level, buf)


    def publish(self, topic, data="", qos=0):
        if self.is_connected:
            self._client.publish(topic, data, qos)

    def subscribe(self, topic):
        if self._client:
            if isinstance(topic, basestring):
                topic = [topic]
            else:
                assert isinstance(topic, (list, tuple)), topic
                assert len(topic)>0
            self._client.subscribe([(x,2) for x in set(topic)])

    def unsubscribe(self, topic):
        if self._client:
            if isinstance(topic, basestring):
                topic = [topic]
            else:
                assert isinstance(topic, (list, tuple)), topic
                assert len(topic)>0
            self._client.unsubscribe([(x,2) for x in set(topic)])


    def get_client_id(self):
        return self._node._key

    def get_will(self):
        return ("{}{}/r".format(NX_PREFIX, self._node.get_id()), '{"ack":"bye"}', 2)

    def make_connection(self, conn):
        conn.reinitialise(client_id=self.get_client_id(), clean_session=True)

        # mqtt callbacks
        conn.on_connect = self._on_connect
        conn.on_disconnect = self._on_disconnect
        conn.on_subscribe = self._on_subscribe
        conn.on_message = self._on_message
        conn.on_log = self._on_log

        # user account
        if self._username!="":
            conn.username_pw_set(self._username, self._password)

        if self._certfile not in (None, ""):
            conn.tls_set(self._cafile, certfile=self._certfile, keyfile=self._keyfile)
            conn.tls_insecure_set(True)

        will = self.get_will()
        if will != None:
            if isinstance(will, basestring):
                will = [will]
            assert isinstance(will, (list, tuple)), will

            wt = will[0]
            wp = None
            wq = 0
            if len(will)>1:
                wp = will[1]

            if len(will)>2:
                wq = will[2]
                assert wq in (0,1,2), wq

            if wt != "":
                conn.will_clear()
                conn.will_set(wt, payload=wp, qos=wq)

        conn.message_callback_add("{}+/q".format(NX_PREFIX), self.ack_msg_request)
        conn.message_callback_add("{}+".format(NX_PREFIX), self.ack_msg_packet)
        return conn


    def connect(self):
        assert self._host!=""
        assert self._node.get_id() > 0

        # reset next_conn_time flag
        self._next_connect_time = 0

        if not self.is_connected:
            self._reconnect_try += 1

            if self._client == None:
                self._client = self.make_connection(Client())
                reconn = False
            else:
                reconn = True

            try:
                if not reconn:
                    if self._port<=0:
                        if self._certfile!="":
                            p = 8883
                        else:
                            p = 1883
                    else:
                        p = self._port
                    self._client.connect(self._host, port=p, keepalive=60)
                else:
                    self._client.reconnect()
            except:
                traceback.print_exc()
                self._sched_next_connect()

        return self._client


    def loop(self):
        if not self.is_connected and (self._next_connect_time>0 and time.time()>=self._next_connect_time):
            self.connect()
        else:
            # drive mqtt socket work
            if self._client:
                self._client.loop(0.002)




#coding: utf-8
from thing import Thing
from UserDict import IterableUserDict
import sqlite3
import sys
import inspect
import os
import zipfile
import traceback

NODEWOX_HOME = "/var/lib/nodewox"

class PackageLoader(object):
    "read thing class from a zip file or path"

    def __init__(self, path):
        assert isinstance(path, basestring), path
        assert os.path.exists(path), path

        path = os.path.abspath(path)
        self._path = path

        if os.path.isdir(path):
            self._classes = self._check_path(path)
        elif zipfile.is_zipfile(path):
            self._classes = self._check_zipfile(path)
        else:
            self._classes = {}

    def _check_path(self, path):
        "load Thing classes from a path"
        assert os.path.isdir(path), path
        if not path.endswith("/"): 
            path = "{}/".format(path)

        mods = set()
        for dirname, subnames, filenames in os.walk(path, followlinks=True):
            if "__init__.py" in filenames:
                mods.add(dirname[len(path):])

        sys.path.insert(0, path)
        cc = self._load_classes_from_modules(mods)
        sys.path.pop(0)
        return cc

    def _check_zipfile(self, zfile):
        "load Thing classes from a zip file"
        assert zipfile.is_zipfile(zfile), zfile

        # possible modules -> mods
        z = zipfile.ZipFile(zfile)
        mods = set()
        for n in z.namelist():
            if n.endswith("/__init__.py"):
                p = n[:-len("/__init__.py")]
                if len(p)>0:
                    mods.add(p)

        sys.path.insert(0, zfile)
        cc = self._load_classes_from_modules(mods)
        z.close()
        sys.path.pop(0)
        return cc

    def _load_classes_from_modules(self, mods):
        "load thing classes from a list of module names"
        assert isinstance(mods, (list, tuple, set)), mods
        assert len(mods)>0

        ret = {}
        for path in mods:
            nn = path.split("/")
            try:
                m = __import__(".".join(nn), fromlist=nn[:-1])
                for cname, x in inspect.getmembers(m, inspect.isclass):
                    if issubclass(x, Thing) and x!=Thing:
                        modname = ".".join(nn)
                        index = "{}.{}".format(modname, cname)
                        ret[index] = (modname, x)
            except:
                continue
        return ret

    @property
    def path(self):
        return self._path

    @property
    def classes(self):
        return dict(**self._classes)

    @classmethod
    def list(ctx):
        "list installed packages"
        for dirname, subnames, filenames in os.walk(os.path.join(NODEWOX_PATH, "packages")):
            if "meta" in filenames:
                print dirname


class Storage(object):
    "db init and connect"

    def __init__(self):
        self._dbfile = os.path.join(NODEWOX_HOME, "nodewox.db")
        if not os.path.exists(self._dbfile):
            self._setup()

    def _setup(self):
        conn = sqlite3.connect(self._dbfile)

        conn.execute("""
            CREATE TABLE REMOTE (
                remote_id varchar(100) primary key not null,
                rest_url text not null,
                rest_ca text not null default '',
                pri int not null default 0
            );""")

        conn.execute("""
            CREATE TABLE META (
                meta_id varchar(100) primary key not null,
                module text not null,
                class text not null,
                name text not null,
                pid int,
                path text not null
            );""")

        conn.execute("""
            CREATE TABLE PROFILE (
                profile_id text primary key not null,
                meta_id text not null,
                name text not null,
                token text not null,
                remote_id text not null,
                password text not null default '',
                owner text not null default '',
                iot_cert text not null default '',
                iot_key text not null default '',
                iot_ca text not null default ''
            );""")

        conn.execute("""insert into REMOTE (remote_id, rest_url, rest_ca, pri) values
            ('nodewox.org', 'https://www.nodewox.org/api', '', 1)""")

        # for test
        fh = open("/home/johnray/devel/nodewox/etc/mqtt_ca.crt")
        testca = fh.read()
        fh.close()
        conn.execute("""insert into REMOTE (remote_id, rest_url, rest_ca) values
            ('test', 'https://192.168.0.7:6601/api', '%s')""" % testca)

        conn.commit()
        conn.close()
        sys.stdout.write("create nodewox storage\n")

    @property
    def conn(self):
        return sqlite3.connect(self._dbfile)


class _TableORM(IterableUserDict):
    "base wrapper for data table"
    PKEY = ""
    TABLE = ""

    def __init__(self, **kwargs):
        assert self.PKEY!=""
        assert self.TABLE!=""
        self._id = kwargs[self.PKEY]
        if len(kwargs)==1 and self.PKEY in kwargs:
            self.data = {}
        else:
            self.data = dict(kwargs)

    @property
    def is_empty(self):
        return len(self.data)==0

    def load(self, conn):
        cur = conn.execute("select * from {} where {}=?".format(self.TABLE, self.PKEY), (self._id,))
        self.data = self._cursor_to_dict(cur) or {}

    def exists(self, conn, key):
        cus = conn.execute(
                "select %(f)s from %(t)s where %(f)s=?" % {"t":self.TABLE, "f":self.PKEY}, \
                (self._id,))
        return cus.fetchone()!=None

    def save(self, conn):
        assert len(self.data)>0
        if self._id not in self.data:
            self.data[self.PKEY] = self._id
        else:
            assert self.data[self.PKEY] == self._id, (self.data, self._id)
        if self.exists(conn, self._id):
            sets = ",".join("{}=:{}".format(x,x) for x in self.data)
            s = "update {} set {} where {}=:{}".format(self.TABLE, sets, self.PKEY, self.PKEY)
        else:
            fields = ",".join(self.data.keys())
            vals = ",".join(":%s" % x for x in self.data.keys())
            s = "insert into {} ({}) values ({})".format(self.TABLE, fields, vals)
        conn.execute(s, self.data)

    @classmethod
    def _cursor_to_dict(cls, cursor):
        vals = cursor.fetchone()
        if vals:
            desc = dict((i,x[0]) for i,x in enumerate(cursor.description))
            return dict((desc[i],x) for i,x in enumerate(vals))

    @classmethod
    def _cursor_to_dictlist(cls, cursor):
        res = []
        lst = cursor.fetchall()
        if len(lst)>0:
            desc = dict((i,x[0]) for i,x in enumerate(cursor.description))
            for vals in lst:
                res.append(dict((desc[i],x) for i,x in enumerate(vals)))
        return res

    @classmethod
    def list(cls, conn):
        cur = conn.execute("select * from {} order by {}".format(cls.TABLE, cls.PKEY))
        return [cls(**x) for x in cls._cursor_to_dictlist(cur)]

    @classmethod
    def find(cls, conn, key):
        cur = conn.execute("select * from {} where {}=?".format(cls.TABLE, cls.PKEY), (key,))
        data = cls._cursor_to_dict(cur)
        if data:
            return cls(**data)


class Remote(_TableORM):
    "a remote setting"
    TABLE = "REMOTE"
    PKEY = "remote_id"

    @classmethod
    def primary(self, conn):
        cur = conn.execute("select * from REMOTE where pri=1")
        data = self._cursor_to_dict(cur)
        if data:
            return Remote(**data)


class Meta(_TableORM):
    "a thing meta"
    TABLE = "META"
    PKEY = "meta_id"

    def __init__(self, **kwargs):
        assert "/" not in kwargs['meta_id']
        assert not kwargs['meta_id'].startswith(".")
        assert not kwargs['meta_id'].endswith(".")
        _TableORM.__init__(self, **kwargs)

    def get_class(self):
        "load and return thing class"
        pkg = PackageLoader(self.data['path'])
        if self.data['meta_id'] in pkg.classes:
            return pkg.classes[self.data['meta_id']][1]


class Profile(_TableORM):
    "a thing registry profile"
    TABLE = "PROFILE"
    PKEY = "profile_id"

    def show_info(self, file=sys.stdout):
        file.write("profile : {}\n".format(self.data['profile_id']))
        file.write("thing   : {}\n".format(self.data['meta_id']))
        file.write("token   : {}\n".format(self.data['token']))
        file.write("remote  : {}\n".format(self.data['remote_id']))
        file.write("owner   : {}\n".format(self.data['owner']))

    def make_thing(self, conn):
        "make thing instance according to profile config"
        assert not self.is_empty
        remote = Remote.find(conn, self.data['remote_id'])
        if not remote:
            return

        meta = Meta.find(conn, self.data['meta_id'])
        if not meta:
            return

        cls = meta.get_class()
        if not cls:
            return

        thing = cls(str(self.data['token']), secret=str(self.data['password']))
        thing.set_rest(str(remote['rest_url']), str(remote['rest_ca']))

        args = [None, None, None]
        for i, k in enumerate(("iot_ca", "iot_cert", "iot_key")):
            if self.data[k] not in (None, ""):
                args[i] = str(self.data[k])
        thing.set_tls(*args)
        return thing

    @classmethod
    def list(cls, conn, meta_id):
        cur = conn.execute( \
                "select * from {} where meta_id=? order by {}".format(cls.TABLE, cls.PKEY), \
                (meta_id,))
        return [cls(**x) for x in cls._cursor_to_dictlist(cur)]

